@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary::Validation;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Implementation of the JSON Schema Validation vocabulary
 
-our $VERSION = '0.589';
+our $VERSION = '0.601';
 
 use 5.020;
 use Moo;
@@ -25,28 +25,30 @@ use namespace::clean;
 
 with 'JSON::Schema::Modern::Vocabulary';
 
-sub vocabulary {
+sub vocabulary ($class) {
   'https://json-schema.org/draft/2019-09/vocab/validation' => 'draft2019-09',
   'https://json-schema.org/draft/2020-12/vocab/validation' => 'draft2020-12';
 }
 
-sub evaluation_order { 1 }
+sub evaluation_order ($class) { 1 }
 
 sub keywords ($class, $spec_version) {
   return (
-    qw(type enum const
-      multipleOf maximum exclusiveMaximum minimum exclusiveMinimum
-      maxLength minLength pattern
-      maxItems minItems uniqueItems),
-    $spec_version ne 'draft7' ? qw(maxContains minContains) : (),
+    qw(type enum),
+    $spec_version ne 'draft4' ? 'const' : (),
+    qw(multipleOf maximum exclusiveMaximum minimum exclusiveMinimum
+      maxLength minLength pattern maxItems minItems uniqueItems),
+    $spec_version !~ /^draft[467]$/ ? qw(maxContains minContains) : (),
     qw(maxProperties minProperties required),
-    $spec_version ne 'draft7' ? 'dependentRequired' : (),
+    $spec_version !~ /^draft[467]$/ ? 'dependentRequired' : (),
   );
 }
 
 sub _traverse_keyword_type ($class, $schema, $state) {
   if (is_plain_arrayref($schema->{type})) {
+    # Note: this is not actually in the spec, but the restriction exists in the metaschema
     return E($state, 'type array is empty') if not $schema->{type}->@*;
+
     foreach my $type ($schema->{type}->@*) {
       return E($state, 'unrecognized type "%s"', $type//'<null>')
         if not any { ($type//'') eq $_ } qw(null boolean object array string number integer);
@@ -62,7 +64,7 @@ sub _traverse_keyword_type ($class, $schema, $state) {
 }
 
 sub _eval_keyword_type ($class, $data, $schema, $state) {
-  my $type = get_type($data);
+  my $type = get_type($data, $state->{spec_version} eq 'draft4' ? { legacy_ints => 1 } : ());
   if (is_plain_arrayref($schema->{type})) {
     return 1 if any {
       $type eq $_ or ($_ eq 'number' and $type eq 'integer')
@@ -82,8 +84,7 @@ sub _eval_keyword_type ($class, $data, $schema, $state) {
 }
 
 sub _traverse_keyword_enum ($class, $schema, $state) {
-  return if not assert_keyword_type($state, $schema, 'array');
-  return 1;
+  return assert_keyword_type($state, $schema, 'array');
 }
 
 sub _eval_keyword_enum ($class, $data, $schema, $state) {
@@ -92,16 +93,15 @@ sub _eval_keyword_enum ($class, $data, $schema, $state) {
   return 1 if any { is_equal($data, $_, $s[$idx++] = {%s}) } $schema->{enum}->@*;
   return E($state, 'value does not match'
     .(!(grep $_->{path}, @s) ? ''
-      : ' (differences start '.join(', ', map 'from item #'.$_.' at "'.$s[$_]->{path}.'"', 0..$#s).')'));
+      : ' ('.join('; ', map "from enum $_ at '$s[$_]->{path}': $s[$_]->{error}", 0..$#s).')'));
 }
 
-sub _traverse_keyword_const { 1 }
+sub _traverse_keyword_const ($class, $schema, $state) { 1 }
 
 sub _eval_keyword_const ($class, $data, $schema, $state) {
   my %s = $state->%{qw(scalarref_booleans stringy_numbers)};
   return 1 if is_equal($data, $schema->{const}, \%s);
-  return E($state, 'value does not match'
-    .($s{path} ? ' (differences start at "'.$s{path}.'")' : ''));
+  return E($state, 'value does not match'.($s{path} ? " (at '$s{path}': $s{error})" : ''));
 }
 
 sub _traverse_keyword_multipleOf ($class, $schema, $state) {
@@ -132,43 +132,79 @@ sub _eval_keyword_multipleOf ($class, $data, $schema, $state) {
   return E($state, 'value is not a multiple of %s', sprintf_num($schema->{multipleOf}));
 }
 
-sub _traverse_keyword_maximum { goto \&_assert_number }
+*_traverse_keyword_maximum = \&_assert_number;
 
 sub _eval_keyword_maximum ($class, $data, $schema, $state) {
   return 1 if not is_type('number', $data)
     and not ($state->{stringy_numbers} and is_type('string', $data) and looks_like_number($data));
-  return 1 if 0+$data <= $schema->{maximum};
-  return E($state, 'value is larger than %s', sprintf_num($schema->{maximum}));
+
+  return 1 if 0+$data < $schema->{maximum};
+  if ($state->{spec_version} eq 'draft4' and $schema->{exclusiveMaximum}) {
+    return E($state, 'value is greater than or equal to %s', sprintf_num($schema->{maximum}));
+  }
+  else {
+    return 1 if 0+$data == $schema->{maximum};
+    return E($state, 'value is greater than %s', sprintf_num($schema->{maximum}));
+  }
 }
 
-sub _traverse_keyword_exclusiveMaximum { goto \&_assert_number }
+sub _traverse_keyword_exclusiveMaximum ($class, $schema, $state) {
+  return _assert_number($class, $schema, $state) if $state->{spec_version} ne 'draft4';
+
+  return if not assert_keyword_type($state, $schema, 'boolean');
+  return E($state, 'use of exclusiveMaximum requires the presence of maximum')
+    if not exists $schema->{maximum};
+  return 1;
+}
 
 sub _eval_keyword_exclusiveMaximum ($class, $data, $schema, $state) {
+  # we do the work in maximum for draft4 so we don't generate multiple errors
+  return 1 if $state->{spec_version} eq 'draft4';
+
   return 1 if not is_type('number', $data)
     and not ($state->{stringy_numbers} and is_type('string', $data) and looks_like_number($data));
+
   return 1 if 0+$data < $schema->{exclusiveMaximum};
-  return E($state, 'value is equal to or larger than %s', sprintf_num($schema->{exclusiveMaximum}));
+  return E($state, 'value is greater than or equal to %s', sprintf_num($schema->{exclusiveMaximum}));
 }
 
-sub _traverse_keyword_minimum { goto \&_assert_number }
+*_traverse_keyword_minimum = \&_assert_number;
 
 sub _eval_keyword_minimum ($class, $data, $schema, $state) {
   return 1 if not is_type('number', $data)
     and not ($state->{stringy_numbers} and is_type('string', $data) and looks_like_number($data));
-  return 1 if 0+$data >= $schema->{minimum};
-  return E($state, 'value is smaller than %s', sprintf_num($schema->{minimum}));
+
+  return 1 if 0+$data > $schema->{minimum};
+  if ($state->{spec_version} eq 'draft4' and $schema->{exclusiveMinimum}) {
+    return E($state, 'value is less than or equal to %s', sprintf_num($schema->{minimum}));
+  }
+  else {
+    return 1 if 0+$data == $schema->{minimum};
+    return E($state, 'value is less than %s', sprintf_num($schema->{minimum}));
+  }
 }
 
-sub _traverse_keyword_exclusiveMinimum { goto \&_assert_number }
+sub _traverse_keyword_exclusiveMinimum ($class, $schema, $state) {
+  return _assert_number($class, $schema, $state) if $state->{spec_version} ne 'draft4';
+
+  return if not assert_keyword_type($state, $schema, 'boolean');
+  return E($state, 'use of exclusiveMinimum requires the presence of minimum')
+    if not exists $schema->{minimum};
+  return 1;
+}
 
 sub _eval_keyword_exclusiveMinimum ($class, $data, $schema, $state) {
+  # we do the work in minimum for draft4 so we don't generate multiple errors
+  return 1 if $state->{spec_version} eq 'draft4';
+
   return 1 if not is_type('number', $data)
     and not ($state->{stringy_numbers} and is_type('string', $data) and looks_like_number($data));
+
   return 1 if 0+$data > $schema->{exclusiveMinimum};
-  return E($state, 'value is equal to or smaller than %s', sprintf_num($schema->{exclusiveMinimum}));
+  return E($state, 'value is less than or equal to %s', sprintf_num($schema->{exclusiveMinimum}));
 }
 
-sub _traverse_keyword_maxLength { goto \&_assert_non_negative_integer }
+*_traverse_keyword_maxLength = \&_assert_non_negative_integer;
 
 sub _eval_keyword_maxLength ($class, $data, $schema, $state) {
   return 1 if not is_type('string', $data);
@@ -176,10 +212,9 @@ sub _eval_keyword_maxLength ($class, $data, $schema, $state) {
   return E($state, 'length is greater than %d', $schema->{maxLength});
 }
 
-sub _traverse_keyword_minLength { goto \&_assert_non_negative_integer }
+*_traverse_keyword_minLength = \&_assert_non_negative_integer;
 
 sub _eval_keyword_minLength ($class, $data, $schema, $state) {
-
   return 1 if not is_type('string', $data);
   return 1 if length($data) >= $schema->{minLength};
   return E($state, 'length is less than %d', $schema->{minLength});
@@ -198,7 +233,7 @@ sub _eval_keyword_pattern ($class, $data, $schema, $state) {
   return E($state, 'pattern does not match');
 }
 
-sub _traverse_keyword_maxItems { goto \&_assert_non_negative_integer }
+*_traverse_keyword_maxItems = \&_assert_non_negative_integer;
 
 sub _eval_keyword_maxItems ($class, $data, $schema, $state) {
   return 1 if not is_type('array', $data);
@@ -206,7 +241,7 @@ sub _eval_keyword_maxItems ($class, $data, $schema, $state) {
   return E($state, 'array has more than %d item%s', $schema->{maxItems}, $schema->{maxItems} > 1 ? 's' : '');
 }
 
-sub _traverse_keyword_minItems { goto \&_assert_non_negative_integer }
+*_traverse_keyword_minItems = \&_assert_non_negative_integer;
 
 sub _eval_keyword_minItems ($class, $data, $schema, $state) {
   return 1 if not is_type('array', $data);
@@ -215,8 +250,7 @@ sub _eval_keyword_minItems ($class, $data, $schema, $state) {
 }
 
 sub _traverse_keyword_uniqueItems ($class, $schema, $state) {
-  return if not assert_keyword_type($state, $schema, 'boolean');
-  return 1;
+  return assert_keyword_type($state, $schema, 'boolean');
 }
 
 sub _eval_keyword_uniqueItems ($class, $data, $schema, $state) {
@@ -228,11 +262,11 @@ sub _eval_keyword_uniqueItems ($class, $data, $schema, $state) {
 
 # Note: no effort is made to check if the 'contains' keyword has been disabled via its vocabulary.
 # The evaluation implementation of maxContains and minContains are in the Applicator vocabulary
-sub _traverse_keyword_maxContains { goto \&_assert_non_negative_integer }
+*_traverse_keyword_maxContains = \&_assert_non_negative_integer;
 
-sub _traverse_keyword_minContains { goto \&_assert_non_negative_integer }
+*_traverse_keyword_minContains = \&_assert_non_negative_integer;
 
-sub _traverse_keyword_maxProperties { goto \&_assert_non_negative_integer }
+*_traverse_keyword_maxProperties = \&_assert_non_negative_integer;
 
 sub _eval_keyword_maxProperties ($class, $data, $schema, $state) {
   return 1 if not is_type('object', $data);
@@ -241,7 +275,7 @@ sub _eval_keyword_maxProperties ($class, $data, $schema, $state) {
     $schema->{maxProperties} > 1 ? 'ies' : 'y');
 }
 
-sub _traverse_keyword_minProperties { goto \&_assert_non_negative_integer }
+*_traverse_keyword_minProperties = \&_assert_non_negative_integer;
 
 sub _eval_keyword_minProperties ($class, $data, $schema, $state) {
   return 1 if not is_type('object', $data);
@@ -252,6 +286,7 @@ sub _eval_keyword_minProperties ($class, $data, $schema, $state) {
 
 sub _traverse_keyword_required ($class, $schema, $state) {
   return if not assert_keyword_type($state, $schema, 'array');
+  return E($state, '"required" array is empty') if $state->{spec_version} eq 'draft4' and not $schema->{required}->@*;
   return E($state, '"required" element is not a string')
     if any { !is_type('string', $_) } $schema->{required}->@*;
   return E($state, '"required" values are not unique') if not is_elements_unique($schema->{required});
@@ -303,8 +338,7 @@ sub _eval_keyword_dependentRequired ($class, $data, $schema, $state) {
 }
 
 sub _assert_number ($class, $schema, $state) {
-  return if not assert_keyword_type($state, $schema, 'number');
-  return 1;
+  return assert_keyword_type($state, $schema, 'number');
 }
 
 sub _assert_non_negative_integer ($class, $schema, $state) {
@@ -328,7 +362,7 @@ JSON::Schema::Modern::Vocabulary::Validation - Implementation of the JSON Schema
 
 =head1 VERSION
 
-version 0.589
+version 0.601
 
 =head1 DESCRIPTION
 
@@ -352,6 +386,14 @@ the equivalent Draft 2019-09 keywords, indicated in metaschemas with the URI C<h
 
 the equivalent Draft 7 keywords that correspond to this vocabulary and are formally specified in L<https://datatracker.ietf.org/doc/html/draft-handrews-json-schema-validation-01#section-6>.
 
+=item *
+
+the equivalent Draft 6 keywords that correspond to this vocabulary and are formally specified in L<https://json-schema.org/draft-06/draft-wright-json-schema-validation-01#rfc.section.6>.
+
+=item *
+
+the equivalent Draft 4 keywords that correspond to this vocabulary and are formally specified in L<https://json-schema.org/draft-04/draft-fge-json-schema-validation-00#rfc.section.5>.
+
 =back
 
 =head1 SUPPORT
